mod error;
use error::ScootalooError;

mod config;
pub use config::parse_toml;
use config::Config;

mod mastodon;
pub use mastodon::register;
use mastodon::{build_basic_status, get_mastodon_token};

mod twitter;
use twitter::*;

mod util;
use crate::util::generate_media_ids;

mod state;
pub use state::{init_db, migrate_db};
use state::{read_state, write_state, TweetToToot};

use elefren::{prelude::*, status_builder::StatusBuilder, Language};
use futures::StreamExt;
use log::info;
use rusqlite::Connection;
use std::sync::Arc;
use tokio::{spawn, sync::Mutex};

const DEFAULT_RATE_LIMIT: usize = 4;
const DEFAULT_PAGE_SIZE: i32 = 200;

/// This is where the magic happens
#[tokio::main]
pub async fn run(config: Config) {
    // open the SQLite connection
    let conn = Arc::new(Mutex::new(
        Connection::open(&config.scootaloo.db_path).unwrap_or_else(|e| {
            panic!(
                "Something went wrong when opening the DB {}: {}",
                &config.scootaloo.db_path, e
            )
        }),
    ));

    let mut stream = futures::stream::iter(config.mastodon.into_values())
        .map(|mastodon_config| {
            // calculate Twitter page size
            let page_size = mastodon_config
                .twitter_page_size
                .unwrap_or_else(|| config.twitter.page_size.unwrap_or(DEFAULT_PAGE_SIZE));

            // create temporary value for each task
            let scootaloo_cache_path = config.scootaloo.cache_path.clone();
            let token = get_oauth2_token(&config.twitter);
            let task_conn = conn.clone();

            spawn(async move {
                info!("Starting treating {}", &mastodon_config.twitter_screen_name);
                // retrieve the last tweet ID for the username
                let lconn = task_conn.lock().await;
                let last_tweet_id = read_state(&lconn, &mastodon_config.twitter_screen_name, None)?
                    .map(|r| r.tweet_id);
                drop(lconn);

                // get reversed, curated user timeline
                let feed = get_user_timeline(
                    &mastodon_config.twitter_screen_name,
                    &token,
                    last_tweet_id,
                    page_size,
                )
                .await?;

                // get Mastodon instance
                let mastodon = get_mastodon_token(&mastodon_config);

                for tweet in &feed {
                    info!("Treating Tweet {} inside feed", tweet.id);

                    let lconn = task_conn.lock().await;
                    // initiate the toot_reply_id var and retrieve the corresponding toot_id
                    let toot_reply_id: Option<String> = tweet.in_reply_to_user_id.and_then(|_| {
                        read_state(
                            &lconn,
                            &mastodon_config.twitter_screen_name,
                            tweet.in_reply_to_status_id,
                        )
                        .unwrap_or(None)
                        .map(|s| s.toot_id)
                    });
                    drop(lconn);

                    // build basic status by just yielding text and dereferencing contained urls
                    let mut status_text = build_basic_status(tweet);

                    // building associative media list
                    let (media_url, status_medias) =
                        generate_media_ids(tweet, &scootaloo_cache_path, &mastodon).await;

                    status_text = status_text.replace(&media_url, "");

                    info!("Building corresponding Mastodon status");

                    let mut status_builder = StatusBuilder::new();

                    status_builder.status(&status_text).media_ids(status_medias);

                    // theard if necessary
                    if let Some(i) = toot_reply_id {
                        status_builder.in_reply_to(&i);
                    }

                    // language if any
                    if let Some(l) = &tweet.lang {
                        if let Some(r) = Language::from_639_1(l) {
                            status_builder.language(r);
                        }
                    }

                    // can be activated for test purposes
                    status_builder.visibility(elefren::status_builder::Visibility::Private);

                    let status = status_builder.build()?;

                    let published_status = mastodon.new_status(status)?;
                    // this will return if it cannot publish the status preventing the last_tweet from
                    // being written into db

                    let ttt_towrite = TweetToToot {
                        twitter_screen_name: mastodon_config.twitter_screen_name.clone(),
                        tweet_id: tweet.id,
                        toot_id: published_status.id,
                    };

                    // write the current state (tweet ID and toot ID) to avoid copying it another time
                    let lconn = task_conn.lock().await;
                    write_state(&lconn, ttt_towrite)?;
                    drop(lconn);
                }
                Ok::<(), ScootalooError>(())
            })
        })
        .buffer_unordered(config.scootaloo.rate_limit.unwrap_or(DEFAULT_RATE_LIMIT));

    // launch and wait for every handle
    while let Some(result) = stream.next().await {
        match result {
            Ok(Err(e)) => eprintln!("Error within thread: {}", e),
            Err(e) => eprintln!("Error with thread: {}", e),
            _ => (),
        }
    }
}
